@@ -21,9 +21,12 @@
 
 #include "qmidictlAbout.h"
 #include "qmidictlMainForm.h"
+
 #include "qmidictlOptionsForm.h"
+#include "qmidictlMidiControlForm.h"
 
 #include "qmidictlOptions.h"
+#include "qmidictlMidiControl.h"
 #include "qmidictlUdpDevice.h"
 #include "qmidictlDialStyle.h"
 
@@ -32,6 +35,29 @@
 #include <QMessageBox>
 
 
+// Possible cubic-root optimization.
+// (borrowed from metamerist.com)
+#include <cmath>
+
+static inline float cbrtf2 ( float x )
+{
+#ifdef CONFIG_FLOAT32
+	// Avoid strict-aliasing optimization (gcc -O2).
+	union { float f; int i; } u;
+	u.f = x;
+	u.i = (u.i / 3) + 710235478;
+	return u.f;
+#else
+	return cbrtf(x);
+#endif
+}
+
+static inline float cubef2 ( float x )
+{
+	return x * x * x;
+}
+
+
 //----------------------------------------------------------------------------
 // qmidictlMainForm -- UI wrapper form.
 //
@@ -54,9 +80,9 @@ qmidictlMainForm::qmidictlMainForm (
 	m_pStripStates = NULL;
 	m_iCurrentStripPage = 0;
 
-	// Jog shuttle last known state.
-	m_iJogShuttleDelta = 0;
-	m_iJogShuttleValue = 0;
+	// Jog-wheel last known state.
+	m_iJogWheelDelta = 0;
+	m_iJogWheelValue = 0;
 
 	// Activity LED counters.
 	m_iMidiInLed  = 0;
@@ -65,9 +91,9 @@ qmidictlMainForm::qmidictlMainForm (
 	// Kind of soft-mutex.
 	m_iBusy = 0;
 
-	// Hack the jog-shuttle dial style and palette...
+	// Hack the jog-wheel dial style and palette...
 	m_pDialStyle = new qmidictlDialStyle();
-	m_ui.jogShuttleDial->setStyle(m_pDialStyle);
+	m_ui.jogWheelDial->setStyle(m_pDialStyle);
 
 	// Pseudo-singleton reference setup.
 	g_pMainForm = this;
@@ -79,12 +105,12 @@ qmidictlMainForm::qmidictlMainForm (
 	QObject::connect(m_ui.optionsAction,
 		SIGNAL(triggered(bool)),
 		SLOT(optionsSlot()));
+	QObject::connect(m_ui.configureAction,
+		SIGNAL(triggered(bool)),
+		SLOT(configureSlot()));
 	QObject::connect(m_ui.aboutAction,
 		SIGNAL(triggered(bool)),
 		SLOT(aboutSlot()));
-	QObject::connect(m_ui.aboutQtAction,
-		SIGNAL(triggered(bool)),
-		SLOT(aboutQtSlot()));
 	QObject::connect(m_ui.exitAction,
 		SIGNAL(triggered(bool)),
 		SLOT(exitSlot()));
@@ -97,9 +123,9 @@ qmidictlMainForm::qmidictlMainForm (
 		SIGNAL(clicked()),
 		SLOT(nextStripPageSlot()));
 
-	QObject::connect(m_ui.jogShuttleDial,
+	QObject::connect(m_ui.jogWheelDial,
 		SIGNAL(valueChanged(int)),
-		SLOT(jogShuttleSlot(int)));
+		SLOT(jogWheelSlot(int)));
 
 	QObject::connect(m_ui.resetButton,
 		SIGNAL(clicked()),
@@ -180,7 +206,8 @@ void qmidictlMainForm::setup (void)
 		return;
 
 	if (!m_pUdpDevice->open(pOptions->sInterface, pOptions->iUdpPort)) {
-		QMessageBox::critical(this, tr("Network Inferface Error"),
+		QMessageBox::critical(this,
+			tr("Network Inferface Error - %1").arg(QMIDICTL_TITLE),
 			tr("The network interface could not be established.\n\n"
 			"Please, make sure you have an on-line network connection "
 			"and try again."));
@@ -367,11 +394,63 @@ void qmidictlMainForm::sendMmcLocate ( unsigned long iLocate )
 }
 
 
+void qmidictlMainForm::sendMmcStep ( int iDelta )
+{
+	unsigned char data;
+	if (iDelta < 0) {
+		data = (-iDelta & 0x3f) | 0x40;
+	} else {
+		data = (+iDelta & 0x3f);
+	}
+
+	sendMmcCommand(MMC_STEP, &data, sizeof(data));
+}
+
+
+// MIDI data transmitter helpers.
+void qmidictlMainForm::sendData3 (
+	unsigned char data1, unsigned char data2, unsigned char data3 )
+{
+	unsigned char data[3];
+
+	data[0] = data1;
+	data[1] = data2;
+	data[2] = data3;
+
+	sendData(data, sizeof(data));
+}
+
+
+void qmidictlMainForm::sendData2 (
+	unsigned char data1, unsigned char data2 )
+{
+	unsigned char data[2];
+
+	data[0] = data1;
+	data[1] = data2;
+
+	sendData(data, sizeof(data));
+}
+
+
+void qmidictlMainForm::sendPitchBend ( unsigned short iChannel, int iValue )
+{
+	unsigned char data[3];
+	unsigned short val = (unsigned short) (0x2000 + iValue);
+
+	data[0] = 0xe0 + iChannel;
+	data[1] = (val & 0x007f);
+	data[2] = (val & 0x3f80) >> 7;
+
+	sendData(data, sizeof(data));
+}
+
+
 // Network transmitter.
 void qmidictlMainForm::sendData ( unsigned char *data, unsigned short len )
 {
-#if defined(DEBUG)
-	fprintf(stderr, "SEND:");
+#ifdef CONFIG_DEBUG
+	fprintf(stderr, "sendData:");
 	for(int i = 0; i < len; ++i)
 		fprintf(stderr, " %02X", data[i]);
 	fprintf(stderr, "\n");
@@ -386,6 +465,100 @@ void qmidictlMainForm::sendData ( unsigned char *data, unsigned short len )
 }
 
 
+// Generic command dispatcher.
+void qmidictlMainForm::sendCommand ( int iCommand, int iTrack, int iValue )
+{
+	qmidictlMidiControl *pMidiControl = qmidictlMidiControl::getInstance();
+	if (pMidiControl == NULL)
+		return;
+	
+	qmidictlMidiControl::Command command
+		= qmidictlMidiControl::Command(iCommand);
+	if (!pMidiControl->isCommandMapped(command))
+		return;
+
+	const qmidictlMidiControl::MapKey& key
+		= pMidiControl->commandMap().value(command);
+
+	unsigned short iChannel = key.channel();
+	if (key.isChannelTrack()) {
+		iChannel &= qmidictlMidiControl::TrackParamMask;
+		iChannel += iTrack;
+	}
+
+	unsigned short iParam = key.param();
+	if (key.isParamTrack()) {
+		iParam &= qmidictlMidiControl::TrackParamMask;
+		iParam += iTrack;
+	}
+
+	if (key.isLogarithmic())
+		iValue = int(127.0f * cbrtf2(float(iValue) / 127.0f));
+
+	bool bOn = (iValue > 0);
+	switch (key.type()) {
+	case qmidictlMidiControl::MMC:
+		switch (command) {
+		case qmidictlMidiControl::RST:
+			sendMmcLocate(0);
+			break;
+		case qmidictlMidiControl::REW:
+			sendMmcCommand(bOn ? MMC_REWIND : MMC_STOP);
+			break;
+		case qmidictlMidiControl::STOP:
+			sendMmcCommand(MMC_STOP);
+			break;
+		case qmidictlMidiControl::PLAY:
+			sendMmcCommand(bOn ? MMC_PLAY : MMC_PAUSE);
+			break;
+		case qmidictlMidiControl::REC:
+			sendMmcCommand(bOn ? MMC_RECORD_STROBE : MMC_RECORD_EXIT);
+			break;
+		case qmidictlMidiControl::FFWD:
+			sendMmcCommand(bOn ? MMC_FAST_FORWARD : MMC_STOP);
+			break;
+		case qmidictlMidiControl::JOG_WHEEL:
+			sendMmcStep(iValue);
+			break;
+		case qmidictlMidiControl::TRACK_SOLO:
+			sendMmcMaskedWrite(MMC_TRACK_SOLO, iTrack, bOn);
+			break;
+		case qmidictlMidiControl::TRACK_MUTE:
+			sendMmcMaskedWrite(MMC_TRACK_MUTE, iTrack, bOn);
+			break;
+		case qmidictlMidiControl::TRACK_REC:
+			sendMmcMaskedWrite(MMC_TRACK_RECORD, iTrack, bOn);
+			break;
+		case qmidictlMidiControl::TRACK_VOL: // Not handled here.
+		default:
+			break;
+		}
+		break;
+	case qmidictlMidiControl::NOTE_OFF:
+		sendData3(0x80 + iChannel, iParam, iValue);
+		break;
+	case qmidictlMidiControl::NOTE_ON:
+		sendData3(0x90 + iChannel, iParam, iValue);
+		break;
+	case qmidictlMidiControl::KEY_PRESS:
+		sendData3(0xa0 + iChannel, iParam, iValue);
+		break;
+	case qmidictlMidiControl::CONTROLLER:
+		sendData3(0xb0 + iChannel, iParam, iValue);
+		break;
+	case qmidictlMidiControl::PGM_CHANGE:
+		sendData2(0xc0 + iChannel, iParam);
+		break;
+	case qmidictlMidiControl::CHAN_PRESS:
+		sendData2(0xd0 + iChannel, iValue);
+		break;
+	case qmidictlMidiControl::PITCH_BEND:
+		sendPitchBend(iChannel, iValue);
+		break;
+	}
+}
+
+
 // Mixer strip slots.
 void qmidictlMainForm::stripRecordSlot ( int iStrip, bool bOn )
 {
@@ -393,7 +566,7 @@ void qmidictlMainForm::stripRecordSlot ( int iStrip, bool bOn )
 	if (iStrip >= 0 && iStrip < (4 * m_iStripPages))
 		m_pStripStates[iStrip].record = bOn;
 
-	sendMmcMaskedWrite(MMC_TRACK_RECORD, iStrip, bOn);
+	sendCommand(qmidictlMidiControl::TRACK_REC, iStrip, int(bOn));
 }
 
 
@@ -403,7 +576,7 @@ void qmidictlMainForm::stripMuteSlot ( int iStrip, bool bOn )
 	if (iStrip >= 0 && iStrip < (4 * m_iStripPages))
 		m_pStripStates[iStrip].mute = bOn;
 
-	sendMmcMaskedWrite(MMC_TRACK_MUTE, iStrip, bOn);
+	sendCommand(qmidictlMidiControl::TRACK_MUTE, iStrip, int(bOn));
 }
 
 
@@ -413,7 +586,7 @@ void qmidictlMainForm::stripSoloSlot ( int iStrip, bool bOn )
 	if (iStrip >= 0 && iStrip < (4 * m_iStripPages))
 		m_pStripStates[iStrip].solo = bOn;
 
-	sendMmcMaskedWrite(MMC_TRACK_SOLO, iStrip, bOn);
+	sendCommand(qmidictlMidiControl::TRACK_SOLO, iStrip, int(bOn));
 }
 
 
@@ -423,42 +596,28 @@ void qmidictlMainForm::stripSliderSlot ( int iStrip, int iValue )
 	if (iStrip >= 0 && iStrip < (4 * m_iStripPages))
 		m_pStripStates[iStrip].slider = iValue;
 
-	// Special slider (will send raw controller event...
-	unsigned char data[3];
-
-	data[0] = 0xbf; // Channel 16 controller.
-	data[1] = 0x40 + (iStrip & 0x3f);
-	data[2] = (unsigned char) ((127 * iValue) / 100);
-
-	sendData(data, sizeof(data));
+	sendCommand(qmidictlMidiControl::TRACK_VOL, iStrip, (127 * iValue) / 100);
 }
 
 
 // Jog wheel slot.
-void qmidictlMainForm::jogShuttleSlot ( int iValue )
+void qmidictlMainForm::jogWheelSlot ( int iValue )
 {
-	int iDelta = (iValue - m_iJogShuttleValue);
-	if ((iDelta * m_iJogShuttleDelta) < 0)
-		iDelta = -(m_iJogShuttleDelta);
+	int iDelta = (iValue - m_iJogWheelValue);
+	if ((iDelta * m_iJogWheelDelta) < 0)
+		iDelta = -(m_iJogWheelDelta);
 
-	unsigned char data;
-	if (iDelta < 0) {
-		data = (-iDelta & 0x3f) | 0x40;
-	} else {
-		data = (+iDelta & 0x3f);
-	}
+	sendCommand(qmidictlMidiControl::JOG_WHEEL, 0, iDelta);
 
-	sendMmcCommand(MMC_STEP, &data, sizeof(data));
-
-	m_iJogShuttleValue = iValue;
-	m_iJogShuttleDelta = iDelta;
+	m_iJogWheelValue = iValue;
+	m_iJogWheelDelta = iDelta;
 }
 
 
 // Reset action slot
 void qmidictlMainForm::resetSlot (void)
 {
-	sendMmcLocate(0);
+	sendCommand(qmidictlMidiControl::RST);
 }
 
 
@@ -472,7 +631,7 @@ void qmidictlMainForm::rewindSlot ( bool bOn )
 	m_ui.forwardButton->setChecked(false);
 	m_iBusy--;
 
-	sendMmcCommand(bOn ? MMC_REWIND : MMC_STOP);
+	sendCommand(qmidictlMidiControl::REW, 0, int(bOn));
 }
 
 
@@ -482,7 +641,7 @@ void qmidictlMainForm::playSlot ( bool bOn )
 	if (m_iBusy > 0)
 		return;
 
-	sendMmcCommand(bOn ? MMC_PLAY : MMC_PAUSE);
+	sendCommand(qmidictlMidiControl::PLAY, 0, int(bOn));
 }
 
 
@@ -499,7 +658,7 @@ void qmidictlMainForm::stopSlot (void)
 	m_ui.playButton->setChecked(false);
 	m_iBusy--;
 
-	sendMmcCommand(MMC_STOP);
+	sendCommand(qmidictlMidiControl::STOP);
 }
 
 
@@ -509,7 +668,7 @@ void qmidictlMainForm::recordSlot ( bool bOn )
 	if (m_iBusy > 0)
 		return;
 
-	sendMmcCommand(bOn ? MMC_RECORD_STROBE : MMC_RECORD_EXIT);
+	sendCommand(qmidictlMidiControl::REC, 0, int(bOn));
 }
 
 
@@ -523,7 +682,7 @@ void qmidictlMainForm::forwardSlot ( bool bOn )
 	m_ui.rewindButton->setChecked(false);
 	m_iBusy--;
 
-	sendMmcCommand(bOn ? MMC_FAST_FORWARD : MMC_STOP);
+	sendCommand(qmidictlMidiControl::FFWD, 0, int(bOn));
 }
 
 
@@ -535,10 +694,14 @@ void qmidictlMainForm::receiveSlot ( const QByteArray& data )
 
 void qmidictlMainForm::recvData ( unsigned char *data, unsigned short len )
 {
+	qmidictlMidiControl *pMidiControl = qmidictlMidiControl::getInstance();
+	if (pMidiControl == NULL)
+		return;
+
 	m_iBusy++;
 
-#if defined(DEBUG)
-	fprintf(stderr, "RECV:");
+#ifdef CONFIG_DEBUG
+	fprintf(stderr, "recvData:");
 	for(int i = 0; i < len; ++i)
 		fprintf(stderr, " %02X", data[i]);
 	fprintf(stderr, "\n");
@@ -553,53 +716,62 @@ void qmidictlMainForm::recvData ( unsigned char *data, unsigned short len )
 	// Handle immediate incoming MIDI data...
 	int iTracks = (4 * m_iStripPages);
 	int iTrack  = -1;
-	
+
+	qmidictlMidiControl::ControlType ctype
+		= qmidictlMidiControl::ControlType(0);
+	unsigned short iChannel = 0;
+	unsigned short iParam = 0;
+	int iValue = 0;
+
 	// SysEx (actually)...
 	if (data[0] == 0xf0 && data[len - 1] == 0xf7) {
 		// MMC command...
 		unsigned char mmcid = qmidictlOptions::getInstance()->iMmcDevice;
 		if (data[1] == 0x7f && data[3] == 0x06
 			&& (mmcid == 0x7f || data[2] == mmcid)) {
+			ctype = qmidictlMidiControl::MMC;
 			MmcCommand cmd = MmcCommand(data[4]);
 			switch (cmd) {
 			case MMC_STOP:
 			case MMC_RESET:
-				m_ui.rewindButton->setChecked(false);
-				m_ui.playButton->setChecked(false);
-				m_ui.recordButton->setChecked(false);
-				m_ui.forwardButton->setChecked(false);
+				iParam = int(qmidictlMidiControl::STOP);
+				iValue = int(true);
 				break;
 			case MMC_PLAY:
 			case MMC_DEFERRED_PLAY:
-				m_ui.playButton->setChecked(true);
+				iParam = int(qmidictlMidiControl::PLAY);
+				iValue = int(true);
 				break;
 			case MMC_REWIND:
-				m_ui.rewindButton->setChecked(true);
-				m_ui.forwardButton->setChecked(false);
+				iParam = int(qmidictlMidiControl::REW);
+				iValue = int(true);
 				break;
 			case MMC_FAST_FORWARD:
-				m_ui.rewindButton->setChecked(false);
-				m_ui.forwardButton->setChecked(true);
+				iParam = int(qmidictlMidiControl::FFWD);
+				iValue = int(true);
 				break;
 			case MMC_RECORD_STROBE:
-				m_ui.recordButton->setChecked(true);
+				iParam = int(qmidictlMidiControl::REC);
+				iValue = int(true);
 				break;
 			case MMC_RECORD_EXIT:
 			case MMC_RECORD_PAUSE:
-				m_ui.recordButton->setChecked(false);
+				iParam = int(qmidictlMidiControl::REC);
+				iValue = int(false);
 				break;
 			case MMC_PAUSE:
-				m_ui.playButton->setChecked(false);
+				iParam = int(qmidictlMidiControl::PLAY);
+				iValue = int(false);
 				break;
 			case MMC_MASKED_WRITE:
 				if (int(data[5]) > 3) {
 					MmcSubCommand scmd = MmcSubCommand(data[6]);
 					iTrack = (data[7] > 0 ? (data[7] * 7) : 0) - 5;
-					bool bOn = false;
+					iValue = int(false);
 					for (int i = 0; i < 7; ++i) {
 						int iMask = (1 << i);
 						if (data[8] & iMask) {
-							bOn = (data[9] & iMask);
+							iValue = (data[9] & iMask);
 							break;
 						}
 						++iTrack;
@@ -608,13 +780,13 @@ void qmidictlMainForm::recvData ( unsigned char *data, unsigned short len )
 					if (iTrack >= 0 && iTrack < iTracks) {
 						switch (scmd) {
 						case MMC_TRACK_RECORD:
-							m_pStripStates[iTrack].record = bOn;
+							iParam = int(qmidictlMidiControl::TRACK_REC);
 							break;
 						case MMC_TRACK_MUTE:
-							m_pStripStates[iTrack].mute = bOn;
+							iParam = int(qmidictlMidiControl::TRACK_MUTE);
 							break;
 						case MMC_TRACK_SOLO:
-							m_pStripStates[iTrack].solo = bOn;
+							iParam = int(qmidictlMidiControl::TRACK_SOLO);
 							break;
 						default:
 							break;
@@ -626,15 +798,104 @@ void qmidictlMainForm::recvData ( unsigned char *data, unsigned short len )
 				break;
 			}
 		}
+	} else {
+		// Channel event...
+		iChannel = (data[0] & 0x0f);
+		iParam = data[1];
+		if (len > 2)
+			iValue = data[2];
+		// Channel control type...
+		switch (data[0] & 0xf0) {
+		case 0x80:
+			ctype = qmidictlMidiControl::NOTE_OFF;
+			break;
+		case 0x90:
+			ctype = qmidictlMidiControl::NOTE_ON;
+			break;
+		case 0xa0:
+			ctype = qmidictlMidiControl::KEY_PRESS;
+			break;
+		case 0xb0:
+			ctype = qmidictlMidiControl::CONTROLLER;
+			break;
+		case 0xc0:
+			ctype = qmidictlMidiControl::PGM_CHANGE;
+			break;
+		case 0xd0:
+			ctype = qmidictlMidiControl::CHAN_PRESS;			
+			break;
+		case 0xe0:
+			ctype = qmidictlMidiControl::PITCH_BEND;
+			iParam = 0;
+			iValue = ((int(data[2]) << 7) | data[1]) - 0x2000;
+			break;
+		}
 	}
-	else
-	// Channel Controller...
-	if (data[0] == 0xbf && len > 2) {
-		iTrack = int(data[1]) - 0x40;
-		int iValue = (100 * int(data[2])) / 127;
-		// Patch corresponding track/strip state...
-		if (iTrack >= 0 && iTrack < iTracks)
-			m_pStripStates[iTrack].slider = iValue;
+
+	// Lookup the command mapping...
+	qmidictlMidiControl::ControlMap::ConstIterator iter
+		= pMidiControl->find(ctype, iChannel, iParam);
+	if (iter != pMidiControl->controlMap().constEnd()) {
+		const qmidictlMidiControl::MapKey& key = iter.key();
+		qmidictlMidiControl::Command command = iter.value();
+		if (key.isChannelTrack())
+			iTrack  = int(iChannel);
+		else if (key.isParamTrack()) {
+			iTrack  = int(iParam);
+			iTrack -= int(key.param() & qmidictlMidiControl::TrackParamMask);
+		}
+		if (key.isLogarithmic())
+			iValue = int(127.0f * cubef2(float(iValue) / 127.0f));
+	#ifdef CONFIG_DEBUG
+		qDebug("recvData: Command=\"%s\" Type=\"%s\" Track=%d Value=%d",
+			qmidictlMidiControl::textFromCommand(command).toUtf8().constData(),
+			qmidictlMidiControl::textFromType(ctype).toUtf8().constData(),
+			iTrack, iValue);
+	#endif
+		switch (command) {
+		case qmidictlMidiControl::RST:
+			// Nothing to do...
+			break;
+		case qmidictlMidiControl::REW:
+			m_ui.rewindButton->setChecked(true);
+			m_ui.forwardButton->setChecked(false);
+			break;
+		case qmidictlMidiControl::STOP:
+			m_ui.rewindButton->setChecked(false);
+			m_ui.playButton->setChecked(false);
+			m_ui.recordButton->setChecked(false);
+			m_ui.forwardButton->setChecked(false);
+			break;
+		case qmidictlMidiControl::PLAY:
+			m_ui.playButton->setChecked(iValue > 0);
+			break;
+		case qmidictlMidiControl::REC:
+			m_ui.recordButton->setChecked(iValue > 0);
+			break;
+		case qmidictlMidiControl::FFWD:
+			m_ui.rewindButton->setChecked(false);
+			m_ui.forwardButton->setChecked(true);
+			break;
+		case qmidictlMidiControl::JOG_WHEEL:
+			// Nothing to do...
+			break;
+		case qmidictlMidiControl::TRACK_SOLO:
+			if (iTrack >= 0 && iTrack < iTracks)
+				m_pStripStates[iTrack].solo = bool(iValue > 0);
+			break;
+		case qmidictlMidiControl::TRACK_MUTE:
+			if (iTrack >= 0 && iTrack < iTracks)
+				m_pStripStates[iTrack].mute = bool(iValue > 0);
+			break;
+		case qmidictlMidiControl::TRACK_REC:
+			if (iTrack >= 0 && iTrack < iTracks)
+				m_pStripStates[iTrack].record = bool(iValue > 0);
+			break;
+		case qmidictlMidiControl::TRACK_VOL:
+			if (iTrack >= 0 && iTrack < iTracks)
+				m_pStripStates[iTrack].slider = (100 * iValue) / 127;
+			break;
+		}
 	}
 
 	// Update corresponding strip state, when currently visible...
@@ -678,6 +939,13 @@ void qmidictlMainForm::optionsSlot (void)
 }
 
 
+// MIDI control configuration dialog.
+void qmidictlMainForm::configureSlot (void)
+{
+	qmidictlMidiControlForm(this).exec();
+}
+
+
 // About dialog.
 void qmidictlMainForm::aboutSlot (void)
 {
@@ -702,13 +970,6 @@ void qmidictlMainForm::aboutSlot (void)
 }
 
 
-// About Qt dialog.
-void qmidictlMainForm::aboutQtSlot (void)
-{
-	QApplication::aboutQt();
-}
-
-
 // Exit/quit action slot.
 void qmidictlMainForm::exitSlot (void)
 {
